unit GameStream;

interface

uses
  Windows, Messages, Classes, SysUtils, Math;

const
  GP_ID_SIZE = 4;
  GP_CURSOR_MIN = GP_ID_SIZE + 1;

type
  GPIdType = Cardinal;

  TGameStream = class;

  TGamePacket = class(TObject)
  private
    FCursor: Cardinal;
    FData: RawByteString;

    procedure SetData(Data: RawByteString = '');
    function GetId: GPIdType;
    procedure SetId(Id: GPIdType);
  public
    property Id: GPIdType read GetId write SetId;
    property Data: RawByteString read FData write SetData;
    property Cursor: Cardinal read FCursor;

    procedure WriteB(Buf: Pointer; Size: Integer; Index: Integer = 0);
    procedure WriteC(Value: Byte; Index: Integer = 0);
    procedure WriteH(Value: Word; Index: Integer = 0);
    procedure WriteD(Value: Integer; Index: Integer = 0);
    procedure WriteQ(Value: Int64; Index: Integer = 0);
    procedure WriteF(Value: Double; Index: Integer = 0);
    procedure WriteSingle(Value: Single; Index: Integer = 0);
    procedure WriteS(Value: string; Index: Integer = 0);
    function ReadC: Byte;
    function ReadH: Word;
    function ReadD: Integer;
    function ReadQ: Int64;
    function ReadF: Double;
    function ReadS: string;
    function ReadSingle: Single;
    function ReadB(Size: Integer): RawByteString;
    function EndOfPacket(RaiseException: Boolean = True): Boolean;
    procedure Skip(Fields: AnsiString);
    procedure Clear;

    constructor Create;
    destructor Destroy; override;
  end;

  TGamePacketEvent = procedure(Packet: TGamePacket; Sender: TGameStream) of object;
  TGameSendEvent = procedure(Data: RawByteString; Sender: TGameStream) of object;

  TGameStream = class(TObject)
  private
    FRecvBuf: RawByteString;
    Packet: TGamePacket;

  public
    PacketEvent: TGamePacketEvent;
    SendEvent: TGameSendEvent;
    Tag: Integer;
    Master: TObject;
    Game: Boolean;

    procedure Reset;
    procedure Send(Data: RawByteString); overload;
    procedure Send(Packet: TGamePacket); overload;
    function ProcessData(const Data: RawByteString): Boolean;

    constructor Create;
    destructor Destroy; override;
  end;

implementation

constructor TGamePacket.Create;
begin
  inherited Create;

  Clear;
end;

destructor TGamePacket.Destroy;
begin
  // Nothing

  inherited Destroy;
end;

procedure TGamePacket.Clear;
begin
  SetData;
end;

procedure TGamePacket.WriteB(Buf: Pointer; Size: Integer; Index: Integer = 0);
var
  L, MinLen: Integer;
begin
  L := Length(FData);
  if Index < GP_CURSOR_MIN then Index := L + 1;
  MinLen := Index + Size - 1;
  if L < MinLen then SetLength(FData, MinLen);
  Move(Buf^, FData[Index], Size);
end;

procedure TGamePacket.WriteC(Value: Byte; Index: Integer = 0);
begin
  WriteB(@Value, 1, Index);
end;

procedure TGamePacket.WriteH(Value: Word; Index: Integer = 0);
begin
  WriteB(@Value, 2, Index);
end;

procedure TGamePacket.WriteD(Value: Integer; Index: Integer = 0);
begin
  WriteB(@Value, 4, Index);
end;

procedure TGamePacket.WriteQ(Value: Int64; Index: Integer = 0);
begin
  WriteB(@Value, 8, Index);
end;

procedure TGamePacket.WriteF(Value: Double; Index: Integer = 0);
begin
  WriteB(@Value, 8, Index);
end;

procedure TGamePacket.WriteSingle(Value: Single; Index: Integer = 0);
begin
  WriteB(@Value, 4, Index);
end;

procedure TGamePacket.WriteS(Value: string; Index: Integer = 0);
var
  S: RawByteString;
  W: Word;
begin
  S := UTF8Encode(Value);
  W := Length(S);
  S := RawByteString(#0#0) + S;
  Move(W, S[1], 2);
  WriteB(@S[1], W + 2, Index);
end;

function TGamePacket.ReadC: Byte;
begin
  if FCursor < GP_CURSOR_MIN then FCursor := GP_CURSOR_MIN;
  if Length(FData) < FCursor then raise Exception.Create('Error in ReadC!');
  Result := Ord(FData[FCursor]);
  Inc(FCursor);
end;

function TGamePacket.ReadH: Word;
begin
  if FCursor < GP_CURSOR_MIN then FCursor := GP_CURSOR_MIN;
  if Length(FData) < FCursor + 1 then raise Exception.Create('Error in ReadH!');
  Move(FData[FCursor], Result, 2);
  Inc(FCursor, 2);
end;

function TGamePacket.ReadD: Integer;
begin
  if FCursor < GP_CURSOR_MIN then FCursor := GP_CURSOR_MIN;
  if Length(FData) < FCursor + 3 then raise Exception.Create('Error in ReadD!');
  Move(FData[FCursor], Result, 4);
  Inc(FCursor, 4);
end;

function TGamePacket.ReadQ: Int64;
begin
  if FCursor < GP_CURSOR_MIN then FCursor := GP_CURSOR_MIN;
  if Length(FData) < FCursor + 7 then raise Exception.Create('Error in ReadQ!');
  Move(FData[FCursor], Result, 8);
  Inc(FCursor, 8);
end;

function TGamePacket.ReadF: Double;
begin
  if FCursor < GP_CURSOR_MIN then FCursor := GP_CURSOR_MIN;
  if Length(FData) < FCursor + 7 then raise Exception.Create('Error in ReadF!');
  Move(FData[FCursor], Result, 8);
  Inc(FCursor, 8);
end;

function TGamePacket.ReadSingle: Single;
begin
  if FCursor < GP_CURSOR_MIN then FCursor := GP_CURSOR_MIN;
  if Length(FData) < FCursor + 3 then raise Exception.Create('Error in ReadSingle!');
  Move(FData[FCursor], Result, 4);
  Inc(FCursor, 4);
end;

function TGamePacket.ReadS: string;
var
  L, DL: Integer;
  S: AnsiString;
begin
  if FCursor < GP_CURSOR_MIN then FCursor := GP_CURSOR_MIN;
  Result := '';
  DL := Length(FData);
  if DL < FCursor then raise Exception.Create('Error in ReadS!');
  L := Ord(FData[FCursor]);
  if DL < FCursor + L then raise Exception.Create('Error in ReadS!');
  if L < 1 then
    begin
      Result := '';
    end
  else
    begin
      SetLength(S, L - 1);
      Move(FData[FCursor + 1], S[1], L - 1);
      Result := string(S);
    end;

  Inc(FCursor, L + 1);
end;

function TGamePacket.ReadB(Size: Integer): RawByteString;
begin
  if FCursor < GP_CURSOR_MIN then FCursor := GP_CURSOR_MIN;
  if Length(FData) < FCursor + Size - 1 then raise Exception.Create('Error in ReadB!');
  Result := Copy(FData, FCursor, Size);
  Inc(FCursor, Size);
end;

function TGamePacket.EndOfPacket(RaiseException: Boolean = True): Boolean;
var
  N: Integer;
begin
  N := Length(FData) + 1 - FCursor;
  Result := N = 0;
  if (not Result) and RaiseException then raise Exception.Create(Format('EndOfPacket expected (%d)!', [N]));
end;

procedure TGamePacket.Skip(Fields: AnsiString);
var
  I, FI, N: Integer;
  C: AnsiChar;
  T: AnsiString;
begin
  Fields := AnsiString(LowerCase(String(Fields)));
  T := '';
  for FI := 1 to Length(Fields) do
    begin
      C := Fields[FI];
      if C in ['0'..'9'] then
        begin
          T := T + C;
          Continue;
        end;

      if T <> '' then
        begin
          N := StrToInt(String(T));
          T := '';
          if N < 1 then raise Exception.Create('Invalid amount!');
        end
      else
        begin
          N := 1;
        end;

      case C of
        'c': Inc(FCursor, 1 * N);
        'h': Inc(FCursor, 2 * N);
        'd': Inc(FCursor, 4 * N);
        'q': Inc(FCursor, 8 * N);
        'f': Inc(FCursor, 8 * N);
        's': for I := 1 to N do ReadS;
      end;
    end;
end;

procedure TGamePacket.SetData(Data: RawByteString = '');
var
  I: Integer;
begin
  if Length(Data) < GP_ID_SIZE then
    begin
      Data := '';
      for I := 0 to GP_ID_SIZE - 1 do
        begin
          Data := Data + #0;
        end;
    end;
  FData := Data;
  FCursor := GP_CURSOR_MIN;
end;

function TGamePacket.GetId: GPIdType;
begin
  Move(FData[1], Result, GP_ID_SIZE);
end;

procedure TGamePacket.SetId(Id: GPIdType);
begin
  Clear;
  Move(Id, FData[1], GP_ID_SIZE);
end;

constructor TGameStream.Create;
begin
  inherited Create;

  Packet := TGamePacket.Create;
  Reset;
end;

destructor TGameStream.Destroy;
begin
  Packet.Free;

  inherited Destroy;
end;

procedure TGameStream.Reset;
begin
  FRecvBuf := '';
end;

procedure TGameStream.Send(Data: RawByteString);
var
  L, W: Word;
  S: RawByteString;
begin
  if Length(Data) = 0 then Exit;

  if not Assigned(SendEvent) then Exit;

  L := Length(Data);
  W := L + 2;
  SetLength(S, W);
  Move(L, S[1], 2);
  Move(Data[1], S[3], L);

  SendEvent(S, Self);
end;

procedure TGameStream.Send(Packet: TGamePacket);
begin
  if Packet = nil then Exit;
  Send(Packet.Data);
end;

function TGameStream.ProcessData(const Data: RawByteString): Boolean;
var
  S: RawByteString;
  Size: Word;
  L, I: Integer;
begin
  FRecvBuf := FRecvBuf + Data;
  L := Length(FRecvBuf);
  I := 1;
  while L > 2 do
    begin
      Move(FRecvBuf[I], Size, 2);
      Inc(Size, 2);
      if Size < 3 then Exit(False);
      if L < Size then Break;

      S := Copy(FRecvBuf, I + 2, Size - 2);
      
      if Assigned(PacketEvent) then
        begin
          Packet.Data := S;
          PacketEvent(Packet, Self);
        end;

      Inc(I, Size);
      Dec(L, Size);
    end;

  if I > 1 then
    begin
      Delete(FRecvBuf, 1, I - 1);
    end;

  Result := True;
end;

end.
